@@ -185,30 +185,6 @@ static void pktcompression_insert_with_labels(struct pktcompression* pcomp,
 	}
 }
 
-/* calculate length of dname in uncompressed wireformat in buffer */
-static size_t dname_length(const uint8_t* buf, size_t len)
-{
-	size_t l = 0;
-	if(!buf || len == 0)
-		return l;
-	while(len > 0 && buf[0] != 0) {
-		size_t lablen = (size_t)(buf[0]);
-		if( (lablen&0xc0) )
-			return 0; /* the name should be uncompressed */
-		if(lablen+1 > len)
-			return 0; /* should fit in the buffer */
-		l += lablen+1;
-		len -= lablen+1;
-		buf += lablen+1;
-	}
-	if(len == 0)
-		return 0; /* end label should fit in buffer */
-	if(buf[0] != 0)
-		return 0; /* must end in root label */
-	l += 1; /* for the end root label */
-	return l;
-}
-
 /* write a compressed domain name into the packet,
  * returns uncompressed wireformat length,
  * 0 if it does not fit and -1 on failure, bad dname. */
@@ -216,7 +192,7 @@ static int pktcompression_write_dname(struct buffer* packet,
 	struct pktcompression* pcomp, const uint8_t* rr, size_t rrlen)
 {
 	size_t wirelen = 0;
-	size_t dname_len = dname_length(rr, rrlen);
+	size_t dname_len = buf_dname_length(rr, rrlen);
 	if(!rr || rrlen == 0 || dname_len == 0)
 		return 0;
 	while(rrlen > 0 && rr[0] != 0) {
@@ -259,6 +235,118 @@ static int pktcompression_write_dname(struct buffer* packet,
 	return wirelen;
 }
 
+static int ixfr_write_rdata_pkt(struct buffer* packet, uint16_t tp,
+	struct pktcompression* pcomp, const uint8_t* rr, size_t rdlen)
+{
+	const struct nsd_type_descriptor* descriptor = nsd_type_descriptor(tp);
+	size_t i;
+	uint16_t offset; /* The offset in rr. */
+
+	/* The rr points at the start of the rdata of length rdlen.
+	 * This is uncompressed wireformat. */
+
+	if(!descriptor->is_compressible) {
+		if(!buffer_available(packet, rdlen))
+			return 0;
+		buffer_write(packet, rr, rdlen);
+		return 1;
+	}
+
+	/* It is compressible, loop over the fields and write compressed
+	 * domain names, when the rdata has a compressible name. */
+	offset = 0;
+	for(i=0; i < descriptor->rdata.length; i++) {
+		const nsd_rdata_descriptor_type* field =
+			&descriptor->rdata.fields[i];
+		uint16_t field_len = 0;
+		int already_written = 0;
+		if(rdlen == offset && field->is_optional)
+			break; /* There are no more rdata fields. */
+		if(field->calculate_length_uncompressed_wire) {
+			/* Call field length function. */
+			/* This is called with an uncompressed wireformat
+			 * data buffer, instead of the in-memory data buffer.
+			 * For IPSECKEY it does not matter, since it has
+			 * a literal dname storage. */
+			struct domain* domain;
+			int32_t l = field->calculate_length_uncompressed_wire(
+				rdlen, rr, offset, &domain);
+			if(l < 0)
+				return 1; /* attempt to skip malformed rr */
+			field_len = l;
+			if(domain) {
+				/* Treat as uncompressed dname, to be safe. */
+				/* Write as an uncompressed name. */
+				if(!buffer_available(packet,
+					domain_dname(domain)->name_size))
+					return 0;
+				buffer_write(packet,
+					dname_name(domain_dname(domain)),
+					domain_dname(domain)->name_size);
+				already_written = 1;
+			}
+		} else if(field->length >= 0) {
+			field_len = field->length;
+		} else {
+			size_t dlen;
+			int dname_len;
+			switch(field->length) {
+				/* The dnames are stored in uncompressed
+				 * wireformat in the uncompressed wireformat
+				 * string. */
+			case RDATA_COMPRESSED_DNAME:
+				/* Attempt to compress the compressible
+				 * name. */
+				dname_len = pktcompression_write_dname(packet,
+					pcomp, rr+offset, rdlen-offset);
+				if(dname_len == -1)
+					return 1; /* attempt to skip malformed rr */
+				if(dname_len == 0)
+					return 0;
+				field_len = dname_len;
+				already_written = 1;
+				break;
+			case RDATA_UNCOMPRESSED_DNAME:
+			case RDATA_LITERAL_DNAME:
+				/* Write as an uncompressed name. */
+				if(rdlen-offset<1)
+					return 1; /* attempt to skip malformed rr */
+				dlen = buf_dname_length(rr+offset,
+					rdlen-offset);
+				if(dlen == 0)
+					return 1; /* attempt to skip malformed rr */
+				field_len = dlen;
+				break;
+			case RDATA_STRING:
+			case RDATA_BINARY:
+				if(rdlen-offset<1)
+					return 1; /* attempt to skip malformed rr */
+				field_len = ((uint16_t)(rr+offset)[0]) + 1;
+				break;
+			case RDATA_IPSECGATEWAY:
+			case RDATA_AMTRELAY_RELAY:
+				/* This should have called the callback. */
+				return 1; /* attempt to skip malformed rr */
+			case RDATA_REMAINDER:
+				field_len = rdlen - offset;
+				break;
+			default:
+				/* Unknown specialized value. */
+				return 1; /* attempt to skip malformed rr */
+			}
+		}
+		if((size_t)offset+field_len > rdlen)
+			return 1; /* attempt to skip malformed rr */
+		if(!already_written) {
+			if(!buffer_available(packet, field_len))
+				return 0;
+			buffer_write(packet, rr+offset, field_len);
+		}
+		offset += field_len;
+	}
+	return 1;
+}
+
 /* write an RR into the packet with compression for domain names,
  * return 0 and resets position if it does not fit in the packet. */
 static int ixfr_write_rr_pkt(struct query* query, struct buffer* packet,
@@ -270,8 +358,6 @@ static int ixfr_write_rr_pkt(struct query* query, struct buffer* packet,
 	uint16_t tp;
 	int dname_len;
 	size_t rdlen;
-	size_t i;
-	rrtype_descriptor_type* descriptor;
 
 	if(total_added == 0) {
 		size_t oldmaxlen = query->maxlen;
@@ -321,97 +407,11 @@ static int ixfr_write_rr_pkt(struct query* query, struct buffer* packet,
 		return 1; /* attempt to skip this malformed rr, could assert */
 
 	/* rdata */
-	descriptor = rrtype_descriptor_by_type(tp);
-	for(i=0; i<descriptor->maximum; i++) {
-		size_t copy_len = 0;
-		if(rdlen == 0)
-			break;
-
-		switch(rdata_atom_wireformat_type(tp, i)) {
-		case RDATA_WF_COMPRESSED_DNAME:
-			dname_len = pktcompression_write_dname(packet, pcomp,
-				rr, rdlen);
-			if(dname_len == -1)
-				return 1; /* attempt to skip malformed rr */
-			if(dname_len == 0) {
-				buffer_set_position(packet, oldpos);
-				return 0;
-			}
-			rr += dname_len;
-			rdlen -= dname_len;
-			break;
-		case RDATA_WF_UNCOMPRESSED_DNAME:
-		case RDATA_WF_LITERAL_DNAME:
-			copy_len = rdlen;
-			break;
-		case RDATA_WF_BYTE:
-			copy_len = 1;
-			break;
-		case RDATA_WF_SHORT:
-			copy_len = 2;
-			break;
-		case RDATA_WF_LONG:
-			copy_len = 4;
-			break;
-		case RDATA_WF_TEXTS:
-		case RDATA_WF_LONG_TEXT:
-			copy_len = rdlen;
-			break;
-		case RDATA_WF_TEXT:
-		case RDATA_WF_BINARYWITHLENGTH:
-			copy_len = 1;
-			if(rdlen > copy_len)
-				copy_len += rr[0];
-			break;
-		case RDATA_WF_A:
-			copy_len = 4;
-			break;
-		case RDATA_WF_AAAA:
-			copy_len = 16;
-			break;
-		case RDATA_WF_ILNP64:
-			copy_len = 8;
-			break;
-		case RDATA_WF_EUI48:
-			copy_len = EUI48ADDRLEN;
-			break;
-		case RDATA_WF_EUI64:
-			copy_len = EUI64ADDRLEN;
-			break;
-		case RDATA_WF_BINARY:
-			copy_len = rdlen;
-			break;
-		case RDATA_WF_APL:
-			copy_len = (sizeof(uint16_t)    /* address family */
-                                  + sizeof(uint8_t)   /* prefix */
-                                  + sizeof(uint8_t)); /* length */
-			if(copy_len <= rdlen)
-				copy_len += (rr[copy_len-1]&APL_LENGTH_MASK);
-			break;
-		case RDATA_WF_IPSECGATEWAY:
-			copy_len = rdlen;
-			break;
-		case RDATA_WF_SVCPARAM:
-			copy_len = 4;
-			if(copy_len <= rdlen)
-				copy_len += read_uint16(rr+2);
-			break;
-		default:
-			copy_len = rdlen;
-			break;
-		}
-		if(copy_len) {
-			if(!buffer_available(packet, copy_len)) {
-				buffer_set_position(packet, oldpos);
-				return 0;
-			}
-			if(copy_len > rdlen)
-				return 1; /* assert of skip malformed */
-			buffer_write(packet, rr, copy_len);
-			rr += copy_len;
-			rdlen -= copy_len;
-		}
+	if(!ixfr_write_rdata_pkt(packet, tp, pcomp, rr, rdlen)) {
+		buffer_set_position(packet, oldpos);
+		return 0;
 	}
+
 	/* write compressed rdata length */
 	buffer_write_u16_at(packet, rdpos, buffer_position(packet)-rdpos-2);
 	if(total_added == 0) {
@@ -479,14 +479,18 @@ static int parse_qserial(struct buffer* packet, uint32_t* qserial,
 	return 0;
 }
 
+/* get serial from SOA rdata */
+static uint32_t soa_rdata_get_serial(uint8_t* rdata, uint16_t rdlength)
+{
+	if(rdlength < 2*sizeof(void*) /* name ptr */ + 4 /* serial */)
+		return 0;
+	return read_uint32(rdata+2*sizeof(void*));
+}
+
 /* get serial from SOA RR */
 static uint32_t soa_rr_get_serial(struct rr* rr)
 {
-	if(rr->rdata_count < 3)
-		return 0;
-	if(rr->rdatas[2].data[0] < 4)
-		return 0;
-	return read_uint32(&rr->rdatas[2].data[1]);
+	return soa_rdata_get_serial(rr->rdata, rr->rdlength);
 }
 
 /* get the current serial from the zone */
@@ -496,11 +500,7 @@ uint32_t zone_get_current_serial(struct zone* zone)
 		return 0;
 	if(zone->soa_rrset->rr_count == 0)
 		return 0;
-	if(zone->soa_rrset->rrs[0].rdata_count < 3)
-		return 0;
-	if(zone->soa_rrset->rrs[0].rdatas[2].data[0] < 4)
-		return 0;
-	return read_uint32(&zone->soa_rrset->rrs[0].rdatas[2].data[1]);
+	return soa_rr_get_serial(zone->soa_rrset->rrs[0]);
 }
 
 /* iterator over ixfr data. find first element, eg. oldest zone version
@@ -804,8 +804,8 @@ query_state_type query_ixfr(struct nsd *nsd, struct query *query)
 			query_add_compression_domain(query, zone->apex,
 				QHEADERSZ);
 			if(packet_encode_rr(query, zone->apex,
-				&zone->soa_rrset->rrs[0],
-				zone->soa_rrset->rrs[0].ttl)) {
+				zone->soa_rrset->rrs[0],
+				zone->soa_rrset->rrs[0]->ttl)) {
 				ANCOUNT_SET(query->packet, 1);
 			} else {
 				RCODE_SET(query->packet, RCODE_SERVFAIL);
@@ -1082,7 +1082,7 @@ void ixfr_store_finish(struct ixfr_store* ixfr_store, struct nsd* nsd,
 }
 
 /* read SOA rdata section for SOA storage */
-static int read_soa_rdata(struct buffer* packet, uint8_t* primns,
+static int read_soa_rdata_fields(struct buffer* packet, uint8_t* primns,
 	int* primns_len, uint8_t* email, int* email_len,
 	uint32_t* serial, uint32_t* refresh, uint32_t* retry,
 	uint32_t* expire, uint32_t* minimum, size_t* sz)
@@ -1171,7 +1171,7 @@ void ixfr_store_add_newsoa(struct ixfr_store* ixfr_store, uint32_t ttl,
 		buffer_set_position(packet, oldpos);
 		return;
 	}
-	if(!read_soa_rdata(packet, primns, &primns_len, email, &email_len,
+	if(!read_soa_rdata_fields(packet, primns, &primns_len, email, &email_len,
 		&serial, &refresh, &retry, &expire, &minimum, &sz)) {
 		log_msg(LOG_ERR, "ixfr_store newsoa: cannot parse packet");
 		ixfr_store_cancel(ixfr_store);
@@ -1225,7 +1225,7 @@ void ixfr_store_add_oldsoa(struct ixfr_store* ixfr_store, uint32_t ttl,
 		buffer_set_position(packet, oldpos);
 		return;
 	}
-	if(!read_soa_rdata(packet, primns, &primns_len, email, &email_len,
+	if(!read_soa_rdata_fields(packet, primns, &primns_len, email, &email_len,
 		&serial, &refresh, &retry, &expire, &minimum, &sz)) {
 		log_msg(LOG_ERR, "ixfr_store oldsoa: cannot parse packet");
 		ixfr_store_cancel(ixfr_store);
@@ -1246,25 +1246,21 @@ void ixfr_store_add_oldsoa(struct ixfr_store* ixfr_store, uint32_t ttl,
 	buffer_set_position(packet, oldpos);
 }
 
-/* store RR in data segment */
-static int ixfr_putrr(const struct dname* dname, uint16_t type, uint16_t klass,
-	uint32_t ttl, rdata_atom_type* rdatas, ssize_t rdata_num,
-	uint8_t** rrs, size_t* rrs_len, size_t* rrs_capacity)
+/* store RR in data segment.
+ * return -1 on fail of wireformat, 0 on allocate failure, or 1 success. */
+static int ixfr_putrr(const rr_type* rr, uint8_t** rrs, size_t* rrs_len,
+	size_t* rrs_capacity)
 {
-	size_t rdlen_uncompressed, sz;
+	int32_t rdlen_uncompressed;
+	size_t sz;
 	uint8_t* sp;
-	int i;
+	const dname_type* dname;
 
-	/* find rdatalen */
-	rdlen_uncompressed = 0;
-	for(i=0; i<rdata_num; i++) {
-		if(rdata_atom_is_domain(type, i)) {
-			rdlen_uncompressed += domain_dname(rdatas[i].domain)
-				->name_size;
-		} else {
-			rdlen_uncompressed += rdatas[i].data[0];
-		}
-	}
+	rdlen_uncompressed = rr_calculate_uncompressed_rdata_length(rr);
+	if (rdlen_uncompressed < 0)
+		return -1; /* malformed */
+
+	dname = domain_dname(rr->owner);
 	sz = dname->name_size + 2 /*type*/ + 2 /*class*/ + 4 /*ttl*/ +
 		2 /*rdlen*/ + rdlen_uncompressed;
 
@@ -1278,36 +1274,18 @@ static int ixfr_putrr(const struct dname* dname, uint16_t type, uint16_t klass,
 	*rrs_len += sz;
 	memmove(sp, dname_name(dname), dname->name_size);
 	sp += dname->name_size;
-	write_uint16(sp, type);
-	sp += 2;
-	write_uint16(sp, klass);
-	sp += 2;
-	write_uint32(sp, ttl);
-	sp += 4;
-	write_uint16(sp, rdlen_uncompressed);
-	sp += 2;
-	for(i=0; i<rdata_num; i++) {
-		if(rdata_atom_is_domain(type, i)) {
-			memmove(sp, dname_name(domain_dname(rdatas[i].domain)),
-				domain_dname(rdatas[i].domain)->name_size);
-			sp += domain_dname(rdatas[i].domain)->name_size;
-		} else {
-			memmove(sp, &rdatas[i].data[1], rdatas[i].data[0]);
-			sp += rdatas[i].data[0];
-		}
-	}
+	write_uint16(sp, rr->type);
+	write_uint16(sp + 2, rr->klass);
+	write_uint32(sp + 4, rr->ttl);
+	write_uint16(sp + 8, rdlen_uncompressed);
+	rr_write_uncompressed_rdata(rr, sp+10, rdlen_uncompressed);
 	return 1;
 }
 
-void ixfr_store_putrr(struct ixfr_store* ixfr_store, const struct dname* dname,
-	uint16_t type, uint16_t klass, uint32_t ttl, struct buffer* packet,
-	uint16_t rrlen, struct region* temp_region, uint8_t** rrs,
-	size_t* rrs_len, size_t* rrs_capacity)
+void ixfr_store_putrr(struct ixfr_store* ixfr_store, const rr_type* rr,
+	uint8_t** rrs, size_t* rrs_len, size_t* rrs_capacity)
 {
-	domain_table_type *temptable;
-	rdata_atom_type *rdatas;
-	ssize_t rdata_num;
-	size_t oldpos;
+	int code;
 
 	if(ixfr_store->cancelled)
 		return;
@@ -1315,7 +1293,7 @@ void ixfr_store_putrr(struct ixfr_store* ixfr_store, const struct dname* dname,
 	/* The SOA data is stored with separate calls. And then appended
 	 * during the finish operation. We do not have to store it here
 	 * when called from difffile's IXFR processing with type SOA. */
-	if(type == TYPE_SOA)
+	if(rr->type == TYPE_SOA)
 		return;
 	/* make space for these RRs we have now; basically once we
 	 * grow beyond the current allowed amount an older IXFR is deleted. */
@@ -1324,72 +1302,57 @@ void ixfr_store_putrr(struct ixfr_store* ixfr_store, const struct dname* dname,
 	if(ixfr_store->cancelled)
 		return;
 
-	/* parse rdata */
-	oldpos = buffer_position(packet);
-	temptable = domain_table_create(temp_region);
-	rdata_num = rdata_wireformat_to_rdata_atoms(temp_region, temptable,
-		type, rrlen, packet, &rdatas);
-	buffer_set_position(packet, oldpos);
-	if(rdata_num == -1) {
-		log_msg(LOG_ERR, "ixfr_store addrr: cannot parse packet");
-		ixfr_store_cancel(ixfr_store);
-		return;
-	}
+	/* store rdata */
+	code = ixfr_putrr(rr, rrs, rrs_len, rrs_capacity);
 
-	if(!ixfr_putrr(dname, type, klass, ttl, rdatas, rdata_num,
-		rrs, rrs_len, rrs_capacity)) {
-		log_msg(LOG_ERR, "ixfr_store addrr: cannot allocate space");
+	if (code <= 0) {
+		if (code == -1)
+			log_msg(LOG_ERR, "ixfr_store addrr: cannot parse rdata format");
+		else
+			log_msg(LOG_ERR, "ixfr_store addrr: cannot allocate space");
 		ixfr_store_cancel(ixfr_store);
 		return;
 	}
 }
 
-void ixfr_store_delrr(struct ixfr_store* ixfr_store, const struct dname* dname,
-	uint16_t type, uint16_t klass, uint32_t ttl, struct buffer* packet,
-	uint16_t rrlen, struct region* temp_region)
+void ixfr_store_delrr(struct ixfr_store* ixfr_store, const rr_type* rr)
 {
-	ixfr_store_putrr(ixfr_store, dname, type, klass, ttl, packet, rrlen,
-		temp_region, &ixfr_store->data->del,
+	if(ixfr_store->cancelled)
+		return;
+	ixfr_store_putrr(ixfr_store, rr, &ixfr_store->data->del,
 		&ixfr_store->data->del_len, &ixfr_store->del_capacity);
 }
 
-void ixfr_store_addrr(struct ixfr_store* ixfr_store, const struct dname* dname,
-	uint16_t type, uint16_t klass, uint32_t ttl, struct buffer* packet,
-	uint16_t rrlen, struct region* temp_region)
+void ixfr_store_addrr(struct ixfr_store* ixfr_store, const rr_type* rr)
 {
-	ixfr_store_putrr(ixfr_store, dname, type, klass, ttl, packet, rrlen,
-		temp_region, &ixfr_store->data->add,
+	if(ixfr_store->cancelled)
+		return;
+	ixfr_store_putrr(ixfr_store, rr, &ixfr_store->data->add,
 		&ixfr_store->data->add_len, &ixfr_store->add_capacity);
 }
 
-int ixfr_store_addrr_rdatas(struct ixfr_store* ixfr_store,
-	const struct dname* dname, uint16_t type, uint16_t klass,
-	uint32_t ttl, rdata_atom_type* rdatas, ssize_t rdata_num)
+int ixfr_store_addrr_rdatas(struct ixfr_store* ixfr_store, const rr_type *rr)
 {
 	if(ixfr_store->cancelled)
 		return 1;
-	if(type == TYPE_SOA)
+	if(rr->type == TYPE_SOA)
 		return 1;
-	return ixfr_putrr(dname, type, klass, ttl, rdatas, rdata_num,
-		&ixfr_store->data->add, &ixfr_store->data->add_len,
-		&ixfr_store->add_capacity);
+	if(ixfr_putrr(rr, &ixfr_store->data->add, &ixfr_store->data->add_len,
+		&ixfr_store->add_capacity) <= 0)
+		return 0;
+	return 1;
 }
 
 int ixfr_store_add_newsoa_rdatas(struct ixfr_store* ixfr_store,
-	const struct dname* dname, uint16_t type, uint16_t klass,
-	uint32_t ttl, rdata_atom_type* rdatas, ssize_t rdata_num)
+	const rr_type* rr)
 {
 	size_t capacity = 0;
-	uint32_t serial;
 	if(ixfr_store->cancelled)
 		return 1;
-	if(rdata_num < 2 || rdata_atom_size(rdatas[2]) < 4)
+	if(!retrieve_soa_rdata_serial(rr, &ixfr_store->data->newserial))
 		return 0;
-	memcpy(&serial, rdata_atom_data(rdatas[2]), sizeof(serial));
-	ixfr_store->data->newserial = ntohl(serial);
-	if(!ixfr_putrr(dname, type, klass, ttl, rdatas, rdata_num,
-		&ixfr_store->data->newsoa, &ixfr_store->data->newsoa_len,
-		&ixfr_store->add_capacity))
+	if(ixfr_putrr(rr, &ixfr_store->data->newsoa,
+		&ixfr_store->data->newsoa_len, &ixfr_store->add_capacity) <= 0)
 		return 0;
 	ixfr_trim_capacity(&ixfr_store->data->newsoa,
 		&ixfr_store->data->newsoa_len, &capacity);
@@ -1464,27 +1427,26 @@ int ixfr_store_oldsoa_uncompressed(struct ixfr_store* ixfr_store,
 	uint8_t* dname, size_t dname_len, uint16_t type, uint16_t klass,
 	uint32_t ttl, uint8_t* rdata, size_t rdata_len)
 {
-	size_t capacity = 0;
+	uint32_t serial;
+	size_t capacity = 0, index, count;
 	if(ixfr_store->cancelled)
 		return 1;
 	if(!ixfr_storerr_uncompressed(dname, dname_len, type, klass,
 		ttl, rdata, rdata_len, &ixfr_store->data->oldsoa,
 		&ixfr_store->data->oldsoa_len, &capacity))
 		return 0;
-	{
-		uint32_t serial;
-		size_t index, count = 0;
-		if (!(count = skip_dname(rdata, rdata_len)))
-			return 0;
-		index = count;
-		if (!(count = skip_dname(rdata+index, rdata_len-index)))
-			return 0;
-		index += count;
-		if (rdata_len - index < 4)
-			return 0;
-		memcpy(&serial, rdata+index, sizeof(serial));
-		ixfr_store->data->oldserial = ntohl(serial);
-	}
+
+	if (!(count = skip_dname(rdata, rdata_len)))
+		return 0;
+	index = count;
+	if (!(count = skip_dname(rdata+index, rdata_len-index)))
+		return 0;
+	index += count;
+	if (rdata_len - index < 4)
+		return 0;
+	memcpy(&serial, rdata+index, sizeof(serial));
+	ixfr_store->data->oldserial = ntohl(serial);
+
 	ixfr_trim_capacity(&ixfr_store->data->oldsoa,
 		&ixfr_store->data->oldsoa_len, &capacity);
 	return 1;
@@ -1971,45 +1933,20 @@ static int ixfr_write_file_header(struct zone* zone, struct ixfr_data* data,
 	return 1;
 }
 
-/* print rdata on one line */
-static int
-oneline_print_rdata(buffer_type *output, rrtype_descriptor_type *descriptor,
-	rr_type* record)
-{
-	size_t i;
-	size_t saved_position = buffer_position(output);
-
-	for (i = 0; i < record->rdata_count; ++i) {
-		if (i == 0) {
-			buffer_printf(output, "\t");
-		} else {
-			buffer_printf(output, " ");
-		}
-		if (!rdata_atom_to_string(
-			    output,
-			    (rdata_zoneformat_type) descriptor->zoneformat[i],
-			    record->rdatas[i], record))
-		{
-			buffer_set_position(output, saved_position);
-			return 0;
-		}
-	}
-
-	return 1;
-}
-
 /* parse wireformat RR into a struct RR in temp region */
 static int parse_wirerr_into_temp(struct zone* zone, char* fname,
 	struct region* temp, uint8_t* buf, size_t len,
-	const dname_type** dname, struct rr* rr)
+	const dname_type** dname, struct rr** rr)
 {
 	size_t bufpos = 0;
-	uint16_t rdlen;
-	ssize_t rdata_num;
+	uint16_t rdlen, tp, klass;
+	uint32_t ttl;
+	int32_t code;
+	const struct nsd_type_descriptor *descriptor;
 	buffer_type packet;
 	domain_table_type* owners;
+	struct domain *domain;
 	owners = domain_table_create(temp);
-	memset(rr, 0, sizeof(*rr));
 	*dname = dname_make(temp, buf, 1);
 	if(!*dname) {
 		log_msg(LOG_ERR, "failed to write zone %s IXFR data %s: failed to parse dname", zone->opts->name, fname);
@@ -2020,11 +1957,11 @@ static int parse_wirerr_into_temp(struct zone* zone, char* fname,
 		log_msg(LOG_ERR, "failed to write zone %s IXFR data %s: buffer too short", zone->opts->name, fname);
 		return 0;
 	}
-	rr->type = read_uint16(buf+bufpos);
+	tp = read_uint16(buf+bufpos);
 	bufpos += 2;
-	rr->klass = read_uint16(buf+bufpos);
+	klass = read_uint16(buf+bufpos);
 	bufpos += 2;
-	rr->ttl = read_uint32(buf+bufpos);
+	ttl = read_uint32(buf+bufpos);
 	bufpos += 4;
 	rdlen = read_uint16(buf+bufpos);
 	bufpos += 2;
@@ -2032,14 +1969,20 @@ static int parse_wirerr_into_temp(struct zone* zone, char* fname,
 		log_msg(LOG_ERR, "failed to write zone %s IXFR data %s: buffer too short for rdatalen", zone->opts->name, fname);
 		return 0;
 	}
+	domain = domain_table_insert(owners, *dname);
 	buffer_create_from(&packet, buf+bufpos, rdlen);
-	rdata_num = rdata_wireformat_to_rdata_atoms(
-		temp, owners, rr->type, rdlen, &packet, &rr->rdatas);
-	if(rdata_num == -1) {
-		log_msg(LOG_ERR, "failed to write zone %s IXFR data %s: cannot parse rdata", zone->opts->name, fname);
+	descriptor = nsd_type_descriptor(tp);
+	code = descriptor->read_rdata(owners, rdlen, &packet, rr);
+	if(code < 0) {
+		log_msg(LOG_ERR, "failed to write zone %s IXFR data %s: cannot parse rdata %s %s %s", zone->opts->name, fname,
+			dname_to_string(*dname,0), rrtype_to_string(tp),
+			read_rdata_fail_str(code));
 		return 0;
 	}
-	rr->rdata_count = rdata_num;
+	(*rr)->owner = domain;
+	(*rr)->type = tp;
+	(*rr)->klass = klass;
+	(*rr)->ttl = ttl;
 	return 1;
 }
 
@@ -2048,16 +1991,14 @@ static int parse_wirerr_into_temp(struct zone* zone, char* fname,
 static int print_rr_oneline(struct buffer* rr_buffer, const dname_type* dname,
 	struct rr* rr)
 {
-	rrtype_descriptor_type *descriptor;
-	descriptor = rrtype_descriptor_by_type(rr->type);
+	const nsd_type_descriptor_type *descriptor = nsd_type_descriptor(
+		rr->type);
 	buffer_printf(rr_buffer, "%s", dname_to_string(dname, NULL));
 	buffer_printf(rr_buffer, "\t%lu\t%s\t%s", (unsigned long)rr->ttl,
 		rrclass_to_string(rr->klass), rrtype_to_string(rr->type));
-	if(!oneline_print_rdata(rr_buffer, descriptor, rr)) {
-		if(!rdata_atoms_to_unknown_string(rr_buffer,
-			descriptor, rr->rdata_count, rr->rdatas)) {
+	if (!print_rdata(rr_buffer, descriptor, rr)) {
+		if(!print_unknown_rdata(rr_buffer, descriptor, rr))
 			return 0;
-		}
 	}
 	return 1;
 }
@@ -2067,7 +2008,7 @@ static int ixfr_write_rr(struct zone* zone, FILE* out, char* fname,
 	uint8_t* buf, size_t len, struct region* temp, buffer_type* rr_buffer)
 {
 	const dname_type* dname;
-	struct rr rr;
+	struct rr* rr;
 
 	if(!parse_wirerr_into_temp(zone, fname, temp, buf, len, &dname, &rr)) {
 		region_free_all(temp);
@@ -2075,7 +2016,7 @@ static int ixfr_write_rr(struct zone* zone, FILE* out, char* fname,
 	}
 
 	buffer_clear(rr_buffer);
-	if(!print_rr_oneline(rr_buffer, dname, &rr)) {
+	if(!print_rr_oneline(rr_buffer, dname, rr)) {
 		log_msg(LOG_ERR, "failed to write zone %s IXFR data %s: cannot spool RR string into buffer", zone->opts->name, fname);
 		region_free_all(temp);
 		return 0;
@@ -2243,6 +2184,11 @@ void ixfr_write_to_file(struct zone* zone, const char* zfile)
 static void domain_table_delete(struct domain_table* table,
 	struct domain* domain)
 {
+	/* first adjust the number list so that domain is the last one */
+	numlist_make_last(table, domain);
+	/* pop off the domain from the number list */
+	(void)numlist_pop_last(table);
+
 #ifdef USE_RADIX_TREE
 	radix_delete(table->nametree, domain->rnode);
 #else
@@ -2270,10 +2216,10 @@ static int can_del_temp_domain(struct domain* domain)
 
 /* delete temporary domain */
 static void ixfr_temp_deldomain(struct domain_table* temptable,
-	struct domain* domain)
+	struct domain* domain, struct domain* avoid)
 {
 	struct domain* p;
-	if(!can_del_temp_domain(domain))
+	if(domain == avoid || !can_del_temp_domain(domain))
 		return;
 	p = domain->parent;
 	/* see if this domain is someones wildcard-child-closest-match,
@@ -2286,7 +2232,7 @@ static void ixfr_temp_deldomain(struct domain_table* temptable,
 	domain_table_delete(temptable, domain);
 	while(p) {
 		struct domain* up = p->parent;
-		if(!can_del_temp_domain(p))
+		if(p == avoid || !can_del_temp_domain(p))
 			break;
 		if(p->parent && p->parent->wildcard_child_closest_match == p)
 			p->parent->wildcard_child_closest_match =
@@ -2300,49 +2246,32 @@ static void ixfr_temp_deldomain(struct domain_table* temptable,
 static void clear_temp_table_of_rr(struct domain_table* temptable,
 	struct zone* tempzone, struct rr* rr)
 {
-#if 0 /* clear out by removing everything, alternate for the cleanout code */
-	/* clear domains from the tempzone,
-	 * the only domain left is the zone apex and its parents */
-	domain_type* domain;
-#ifdef USE_RADIX_TREE
-	struct radnode* first = radix_first(temptable->nametree);
-	domain = first?(domain_type*)first->elem:NULL;
-#else
-	domain = (domain_type*)rbtree_first(temptable->names_to_domains);
-#endif
-	while(domain != (domain_type*)RBTREE_NULL && domain) {
-		domain_type* next = domain_next(domain);
-		if(domain != tempzone->apex &&
-			!domain_is_subdomain(tempzone->apex, domain)) {
-			domain_table_delete(temptable, domain);
-		} else {
-			if(!domain->parent /* is the root */ ||
-				domain == tempzone->apex)
-				domain->usage = 1;
-			else	domain->usage = 0;
-		}
-		domain = next;
-	}
-
-	if(rr->owner == tempzone->apex) {
-		tempzone->apex->rrsets = NULL;
-		tempzone->soa_rrset = NULL;
-		tempzone->soa_nx_rrset = NULL;
-		tempzone->ns_rrset = NULL;
-	}
-	return;
-#endif
+	const nsd_type_descriptor_type* descriptor =
+		nsd_type_descriptor(rr->type);
 
 	/* clear domains in the rdata */
-	unsigned i;
-	for(i=0; i<rr->rdata_count; i++) {
-		if(rdata_atom_is_domain(rr->type, i)) {
-			/* clear out that dname */
-			struct domain* domain =
-				rdata_atom_domain(rr->rdatas[i]);
-			domain->usage --;
-			if(domain != tempzone->apex && domain->usage == 0)
-				ixfr_temp_deldomain(temptable, domain);
+	if(descriptor->has_references) {
+		uint16_t offset = 0;
+		size_t i;
+		for(i=0; i < descriptor->rdata.length; i++) {
+			uint16_t field_len;
+			struct domain* domain;
+			if(rr->rdlength == offset &&
+				descriptor->rdata.fields[i].is_optional)
+				break; /* There are no more rdata fields. */
+			if(!lookup_rdata_field_entry(descriptor, i, rr, offset,
+				&field_len, &domain))
+				break; /* malformed */
+			if(domain != NULL) {
+				/* The field is a domain reference. */
+				/* clear out that dname */
+				domain->usage --;
+				if(domain != tempzone->apex &&
+					domain->usage == 0)
+					ixfr_temp_deldomain(temptable, domain,
+						rr->owner);
+			}
+			offset += field_len;
 		}
 	}
 
@@ -2355,7 +2284,7 @@ static void clear_temp_table_of_rr(struct domain_table* temptable,
 	} else {
 		rr->owner->rrsets = NULL;
 		if(rr->owner->usage == 0) {
-			ixfr_temp_deldomain(temptable, rr->owner);
+			ixfr_temp_deldomain(temptable, rr->owner, NULL);
 		}
 	}
 }
@@ -2367,6 +2296,7 @@ static int ixfr_data_readnewsoa(struct ixfr_data* data, struct zone* zone,
 	uint32_t dest_serial)
 {
 	size_t capacity = 0;
+	int code;
 	if(rr->type != TYPE_SOA) {
 		zone_error(parser, "zone %s ixfr data: IXFR data does not start with SOA",
 			zone->opts->name);
@@ -2394,9 +2324,13 @@ static int ixfr_data_readnewsoa(struct ixfr_data* data, struct zone* zone,
 			dest_serial);
 		return 0;
 	}
-	if(!ixfr_putrr(domain_dname(rr->owner), rr->type, rr->klass, rr->ttl, rr->rdatas, rr->rdata_count, &data->newsoa, &data->newsoa_len, &capacity)) {
-		zone_error(parser, "zone %s ixfr data: cannot allocate space",
-			zone->opts->name);
+	if((code=ixfr_putrr(rr, &data->newsoa, &data->newsoa_len, &capacity))
+		<= 0) {
+		if(code == -1)
+			zone_error(parser, "zone %s ixfr data: cannot parse rdata format",
+				zone->opts->name);
+		else zone_error(parser, "zone %s ixfr data: cannot allocate space",
+				zone->opts->name);
 		return 0;
 	}
 	clear_temp_table_of_rr(temptable, tempzone, rr);
@@ -2433,7 +2367,7 @@ static int ixfr_data_readoldsoa(struct ixfr_data* data, struct zone* zone,
 		return 0;
 	}
 	data->oldserial = soa_rr_get_serial(rr);
-	if(!ixfr_putrr(domain_dname(rr->owner), rr->type, rr->klass, rr->ttl, rr->rdatas, rr->rdata_count, &data->oldsoa, &data->oldsoa_len, &capacity)) {
+	if(!ixfr_putrr(rr, &data->oldsoa, &data->oldsoa_len, &capacity)) {
 		zone_error(parser, "zone %s ixfr data: cannot allocate space",
 			zone->opts->name);
 		return 0;
@@ -2451,19 +2385,20 @@ static int ixfr_data_readdel(struct ixfr_data* data, struct zone* zone,
 	struct domain_table* temptable, struct zone* tempzone)
 {
 	size_t capacity = 0;
-	if(!ixfr_putrr(domain_dname(rr->owner), rr->type, rr->klass, rr->ttl, rr->rdatas, rr->rdata_count, &data->del, &data->del_len, &capacity)) {
+	if(!ixfr_putrr(rr, &data->del, &data->del_len, &capacity)) {
 		zone_error(parser, "zone %s ixdr data: cannot allocate space",
 			zone->opts->name);
 		return 0;
 	}
 	clear_temp_table_of_rr(temptable, tempzone, rr);
-	region_free_all(tempregion);
 	/* check SOA and also serial, because there could be other
 	 * add and del sections from older versions collated, we can
 	 * see this del section end when it has the serial */
 	if(rr->type != TYPE_SOA && soa_rr_get_serial(rr) != data->newserial) {
+		region_free_all(tempregion);
 		return 1;
 	}
+	region_free_all(tempregion);
 	ixfr_trim_capacity(&data->del, &data->del_len, &capacity);
 	return 2;
 }
@@ -2474,16 +2409,17 @@ static int ixfr_data_readadd(struct ixfr_data* data, struct zone* zone,
 	struct domain_table* temptable, struct zone* tempzone)
 {
 	size_t capacity = 0;
-	if(!ixfr_putrr(domain_dname(rr->owner), rr->type, rr->klass, rr->ttl, rr->rdatas, rr->rdata_count, &data->add, &data->add_len, &capacity)) {
+	if(!ixfr_putrr(rr, &data->add, &data->add_len, &capacity)) {
 		zone_error(parser, "zone %s ixfr data: cannot allocate space",
 			zone->opts->name);
 		return 0;
 	}
 	clear_temp_table_of_rr(temptable, tempzone, rr);
-	region_free_all(tempregion);
 	if(rr->type != TYPE_SOA || soa_rr_get_serial(rr) != data->newserial) {
+		region_free_all(tempregion);
 		return 1;
 	}
+	region_free_all(tempregion);
 	ixfr_trim_capacity(&data->add, &data->add_len, &capacity);
 	return 2;
 }
@@ -2491,7 +2427,8 @@ static int ixfr_data_readadd(struct ixfr_data* data, struct zone* zone,
 struct ixfr_data_state {
 	struct zone *zone;
 	struct ixfr_data *data;
-	struct region *tempregion, *stayregion;
+	struct region *stayregion;
+	struct region *tempregion;
 	struct domain_table *temptable;
 	struct zone *tempzone;
 	uint32_t *dest_serial;
@@ -2513,9 +2450,9 @@ static int32_t ixfr_data_accept(
 	const struct dname *dname;
 	struct domain *domain;
 	struct buffer buffer;
-	union rdata_atom *rdatas;
-	ssize_t rdata_count;
 	struct ixfr_data_state *state = (struct ixfr_data_state *)user_data;
+	const struct nsd_type_descriptor *descriptor;
+	int32_t code;
 
 	assert(parser);
 
@@ -2526,17 +2463,32 @@ static int32_t ixfr_data_accept(
 	domain = domain_table_insert(state->temptable, dname);
 	assert(domain);
 
-	rdata_count = rdata_wireformat_to_rdata_atoms(
-		state->tempregion, state->temptable, type, rdlength, &buffer, &rdatas);
-	assert(rdata_count > 0);
-	rr = region_alloc(state->tempregion, sizeof(*rr));
+	descriptor = nsd_type_descriptor(type);
+	code = descriptor->read_rdata(state->temptable, rdlength, &buffer, &rr);
+	/* This has validated the fields on the rdata. The content can be
+	 * dealt with, if this is successful, later on by iterating over the
+	 * rdata fields. For compression, and for printout, the rdata field
+	 * format is known to be good.
+	 * If the field validation is not needed, the wireformat in the
+	 * rdata, rdlength could have been used to add to the ixfr store.
+	 * But it is more prudent to validate the rdata fields. */
+	if(code < 0) {
+		if(verbosity >= 3) {
+			zone_log(parser, ZONE_ERROR, "the RR rdata fields are wrong for the type");
+		}
+		VERBOSITY(3, (LOG_INFO, "zone %s IXFR bad RR, cannot parse "
+			"rdata of %s %s %s", state->zone->opts->name,
+			dname_to_string(dname, NULL), rrtype_to_string(type),
+			read_rdata_fail_str(code)));
+		if(code == TRUNCATED)
+			return ZONE_OUT_OF_MEMORY;
+		return ZONE_BAD_PARAMETER;
+	}
 	assert(rr);
 	rr->owner = domain;
-	rr->rdatas = rdatas;
 	rr->ttl = ttl;
 	rr->type = type;
 	rr->klass = class;
-	rr->rdata_count = rdata_count;
 
 	if (state->rr_count == 0) {
 		if (!ixfr_data_readnewsoa(state->data, state->zone, rr, parser,
