@@ -36,6 +36,7 @@
 #include "options.h"
 #include "nsec3.h"
 #include "tsig.h"
+#include "rdata.h"
 
 /* [Bug #253] Adding unnecessary NS RRset may lead to undesired truncation.
  * This function determines if the final response packet needs the NS RRset
@@ -706,7 +707,7 @@ struct additional_rr_types rt_additional_rr_types[] = {
 
 static void
 add_additional_rrsets(struct query *query, answer_type *answer,
-		      rrset_type *master_rrset, size_t rdata_index,
+		      rrset_type *master_rrset, uint16_t rdata_offset,
 		      int allow_glue, struct additional_rr_types types[])
 {
 	size_t i;
@@ -714,11 +715,11 @@ add_additional_rrsets(struct query *query, answer_type *answer,
 	assert(query);
 	assert(answer);
 	assert(master_rrset);
-	assert(rdata_atom_is_domain(rrset_rrtype(master_rrset), rdata_index));
 
 	for (i = 0; i < master_rrset->rr_count; ++i) {
 		int j;
-		domain_type *additional = rdata_atom_domain(master_rrset->rrs[i].rdatas[rdata_index]);
+		domain_type *additional = rdata_domain_ref_offset(
+			master_rrset->rrs[i], rdata_offset);
 		domain_type *match = additional;
 
 		assert(additional);
@@ -796,30 +797,40 @@ add_rrset(struct query   *query,
 	case TYPE_NS:
 #if defined(INET6)
 		/* if query over IPv6, swap A and AAAA; put AAAA first */
-		add_additional_rrsets(query, answer, rrset, 0, 1,
+		add_additional_rrsets(query, answer, rrset,
+			0, 1,
 			(query->client_addr.ss_family == AF_INET6)?
 			swap_aaaa_additional_rr_types:
 			default_additional_rr_types);
 #else
-		add_additional_rrsets(query, answer, rrset, 0, 1,
+		add_additional_rrsets(query, answer, rrset,
+				      0, 1,
 				      default_additional_rr_types);
 #endif
 		break;
 	case TYPE_MB:
-		add_additional_rrsets(query, answer, rrset, 0, 0,
+		add_additional_rrsets(query, answer, rrset,
+				      0, 0,
 				      default_additional_rr_types);
 		break;
 	case TYPE_MX:
+		add_additional_rrsets(query, answer, rrset,
+				      2, 0,
+				      default_additional_rr_types);
+		break;
 	case TYPE_KX:
-		add_additional_rrsets(query, answer, rrset, 1, 0,
+		add_additional_rrsets(query, answer, rrset,
+				      2, 0,
 				      default_additional_rr_types);
 		break;
 	case TYPE_RT:
-		add_additional_rrsets(query, answer, rrset, 1, 0,
+		add_additional_rrsets(query, answer, rrset,
+				      2, 0,
 				      rt_additional_rr_types);
 		break;
 	case TYPE_SRV:
-		add_additional_rrsets(query, answer, rrset, 3, 0,
+		add_additional_rrsets(query, answer, rrset,
+				      6, 0,
 				      default_additional_rr_types);
 		break;
 	default:
@@ -856,10 +867,10 @@ query_synthesize_cname(struct query* q, struct answer* answer, const dname_type*
 	for(j=0; j<answer->rrset_count; ++j) {
 		if(answer->section[j] == ANSWER_SECTION &&
 			answer->rrsets[j]->rr_count == 1 &&
-			answer->rrsets[j]->rrs[0].type == TYPE_CNAME &&
-			dname_compare(domain_dname(answer->rrsets[j]->rrs[0].owner), from_name) == 0 &&
-			answer->rrsets[j]->rrs[0].rdata_count == 1 &&
-			dname_compare(domain_dname(answer->rrsets[j]->rrs[0].rdatas->domain), to_name) == 0) {
+			answer->rrsets[j]->rrs[0]->type == TYPE_CNAME &&
+			dname_compare(domain_dname(answer->rrsets[j]->rrs[0]->owner), from_name) == 0 &&
+			answer->rrsets[j]->rrs[0]->rdlength >= 1 &&
+			dname_compare(domain_dname(rdata_domain_ref(answer->rrsets[j]->rrs[0])), to_name) == 0) {
 			DEBUG(DEBUG_QUERY,2, (LOG_INFO, "loop for synthesized CNAME rrset for query %s", dname_to_string(q->qname, NULL)));
 			return 0;
 		}
@@ -915,20 +926,26 @@ query_synthesize_cname(struct query* q, struct answer* answer, const dname_type*
 	*to_closest_match = cname_dest;
 
 	/* allocate the CNAME RR */
-	rrset = (rrset_type*) region_alloc(q->region, sizeof(rrset_type));
+	rrset = (rrset_type*) region_alloc(q->region, sizeof(rrset_type)
+#ifdef PACKED_STRUCTS
+		+ sizeof(rr_type*)
+#endif
+		);
 	memset(rrset, 0, sizeof(rrset_type));
 	rrset->zone = q->zone;
 	rrset->rr_count = 1;
-	rrset->rrs = (rr_type*) region_alloc(q->region, sizeof(rr_type));
-	memset(rrset->rrs, 0, sizeof(rr_type));
-	rrset->rrs->owner = cname_domain;
-	rrset->rrs->ttl = ttl;
-	rrset->rrs->type = TYPE_CNAME;
-	rrset->rrs->klass = CLASS_IN;
-	rrset->rrs->rdata_count = 1;
-	rrset->rrs->rdatas = (rdata_atom_type*)region_alloc(q->region,
-		sizeof(rdata_atom_type));
-	rrset->rrs->rdatas->domain = cname_dest;
+#ifndef PACKED_STRUCTS
+	rrset->rrs = (rr_type**) region_alloc(q->region, sizeof(rr_type*));
+#endif
+	rrset->rrs[0] = (rr_type*) region_alloc(q->region,
+		sizeof(rr_type)+sizeof(void*));
+	memset(rrset->rrs[0], 0, sizeof(rr_type));
+	rrset->rrs[0]->owner = cname_domain;
+	rrset->rrs[0]->ttl = ttl;
+	rrset->rrs[0]->type = TYPE_CNAME;
+	rrset->rrs[0]->klass = CLASS_IN;
+	rrset->rrs[0]->rdlength = sizeof(void*);
+	memcpy(rrset->rrs[0]->rdata, &cname_dest, sizeof(void*));
 
 	if(!add_rrset(q, answer, ANSWER_SECTION, cname_domain, rrset)) {
 		DEBUG(DEBUG_QUERY,2, (LOG_INFO, "could not add synthesized CNAME rrset to packet for query %s", dname_to_string(q->qname, NULL)));
@@ -1116,7 +1133,7 @@ answer_domain(struct nsd* nsd, struct query *q, answer_type *answer,
 		assert(rrset->rr_count > 0);
 		if (added) {
 			/* only process first CNAME record */
-			domain_type *closest_match = rdata_atom_domain(rrset->rrs[0].rdatas[0]);
+			domain_type *closest_match = rdata_domain_ref(rrset->rrs[0]);
 			domain_type *closest_encloser = closest_match;
 			zone_type* origzone = q->zone;
 			++q->cname_count;
@@ -1185,7 +1202,7 @@ answer_authoritative(struct nsd   *nsd,
 		/* process DNAME */
 		const dname_type* name = qname;
 		domain_type* src = closest_encloser;
-		domain_type *dest = rdata_atom_domain(rrset->rrs[0].rdatas[0]);
+		domain_type *dest = rdata_domain_ref(rrset->rrs[0]);
 		const dname_type* newname;
 		size_t newnum = 0;
 		zone_type* origzone = q->zone;
@@ -1219,7 +1236,7 @@ answer_authoritative(struct nsd   *nsd,
 		(void)namedb_lookup(nsd->db, newname, &closest_match, &closest_encloser);
 		/* synthesize CNAME record */
 		newnum = query_synthesize_cname(q, answer, name, newname,
-			src, closest_encloser, &closest_match, rrset->rrs[0].ttl);
+			src, closest_encloser, &closest_match, rrset->rrs[0]->ttl);
 		if(!newnum) {
 			/* could not synthesize the CNAME. */
 			/* return previous CNAMEs to make resolver recurse for us */
@@ -1811,8 +1828,8 @@ query_add_optional(query_type *q, nsd_type *nsd, uint32_t *now_p)
 		if(q->edns.zoneversion
 		&& q->zone
 		&& q->zone->soa_rrset
-		&& q->zone->soa_rrset->rrs
-		&& q->zone->soa_rrset->rrs->rdata_count >= 3)
+		&& q->zone->soa_rrset->rr_count >= 1
+		&& q->zone->soa_rrset->rrs[0]->rdlength >= 20 /* 5x 32bit numbers */ +2*sizeof(void*) /* two pointers to domain names */)
 			q->edns.opt_reserved_space += sizeof(uint16_t)
 			                           +  sizeof(uint16_t)
 			                           +  sizeof(uint8_t)
@@ -1836,8 +1853,9 @@ query_add_optional(query_type *q, nsd_type *nsd, uint32_t *now_p)
 			if(q->edns.zoneversion
 			&& q->zone
 			&& q->zone->soa_rrset
-			&& q->zone->soa_rrset->rrs
-			&& q->zone->soa_rrset->rrs->rdata_count >= 3) {
+			&& q->zone->soa_rrset->rr_count >= 1
+			&& q->zone->soa_rrset->rrs[0]->rdlength >= 20+2*sizeof(void*) /* 5x4 bytes and 2 pointers to domains */ ) {
+				uint32_t serial = 0;
 				buffer_write_u16(q->packet, ZONEVERSION_CODE);
 				buffer_write_u16( q->packet
 				                , sizeof(uint8_t)
@@ -1847,9 +1865,8 @@ query_add_optional(query_type *q, nsd_type *nsd, uint32_t *now_p)
 				    domain_dname(q->zone->apex)->label_count - 1);
 				buffer_write_u8( q->packet
 				               , ZONEVERSION_SOA_SERIAL);
-				buffer_write_u32(q->packet,
-				    read_uint32(rdata_atom_data(
-				    q->zone->soa_rrset->rrs->rdatas[2])));
+				retrieve_soa_rdata_serial(q->zone->soa_rrset->rrs[0], &serial);
+				buffer_write_u32(q->packet, serial);
 			}
 			if(q->edns.cookie_status != COOKIE_NOT_PRESENT) {
 				/* cookie opt header */
